package com.sillycycle.bagleyd.oct;

/*
 * @(#)OctCanvas.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>OctCanvas</code> class does all the graphics for
 * <code>OctFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

/*
Beyond Rubik's Cube: spheres, pyramids, dodecahedrons and God knows what
else by Douglas R. Hofstadter, Scientific American, July 1982, pp 16-31.

Magic Cubes 1996 Catalog of Dr. Christoph Bandelow.
*/

public class OctCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 100;

	static final int TOP = 0;
	static final int TR = 1;
	static final int RIGHT = 2;
	static final int BR = 3;
	static final int BOTTOM = 4;
	static final int BL = 5;
	static final int LEFT = 6;
	static final int TL = 7;
	static final int COORD = 8;
	static final int CW = 9;
	static final int CCW = 15;
	static final int SAME = 0;
	static final int OPPOSITE = 1;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int MAX_VIEWS = 2;
	static final int MAX_FACES = 8;
	static final int MAX_SIDES = (MAX_FACES / MAX_VIEWS);
	static final int MAX_ORIENT = (3 * MAX_SIDES);
	static final int IGNORE = -1;

	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "oct.dat";
	static final String SYMBOL = ":";

	Color       foreground, background;
	Color       borderColor;
	Color       currentForeground;
	Color[]     faceColor = new Color[MAX_FACES];
	int[]       periodOfPosition = null;
	int[]       directionOfPosition = null;
	int[]       startPosition = null;
	int	 currentStyle = -1;
	int	 currentDirection = -1;
	int	 currentFace = -1, nextFace = -1;
	int	 currentPosition = -1, nextPosition = -1;
	int	 currentView = -1, nextView = -1;
	boolean     nextShift = false;
	boolean     currentControl = false, nextControl = false;
	int	 rowPosition;
	boolean     started, cheat, vertical;
	boolean     mono, reverse, orient, sticky, practice;
	int	 delay;
	int	 lastX, lastY, dr;
	int	 size, period, sizeSize;
	Point       coreSize;
	Point       offset, letter;
	Point       puzzleSize, puzzleOffset, offsetD;
	int	 delta, facetLength, faceLength, viewLength, viewMiddle;
	int	 orientLineLength;
	OctRTT rtt = new OctRTT(-1, -1, -1);

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	Random      generator;
	OctStack    undo = null, redo = null;
	OctSolve    solve = null;
	String      stringSave = null;
	String[]    token = null;
	String      faceNames = null;

	static final int[][] directionNextFace = {
		/* coord, face */
		{-1, 0, -1, 0, -1, 4, -1, 4},
		{6, 5, 1, 0, 2, 1, 5, 4},
		{1, -1, 1, -1, 5, -1, 5, -1},
		{1, 5, 4, 2, 5, 1, 0, 6},
		{-1, 2, -1, 2, -1, 6, -1, 6},
		{3, 2, 4, 7, 7, 6, 0, 3},
		{3, -1, 3, -1, 7, -1, 7, -1},
		{6, 0, 3, 7, 2, 4, 7, 3}
	};
	static final int[][] directionNextFaceP3 = {
		/* coord, face */
		{-1, 6, -1, 6, -1, 2, -1, 2},
		{5, 6, 0, 1, 1, 2, 4, 5},
		{5, -1, 5, -1, 1, -1, 1, -1},
		{2, 4, 5, 1, 6, 0, 1, 5},
		{-1, 4, -1, 4, -1, 0, -1, 0},
		{2, 3, 7, 4, 6, 7, 3, 0},
		{7, -1, 7, -1, 3, -1, 3, -1},
		{7, 3, 0, 6, 3, 7, 4, 2}
	};
	OctLoc[][] rowToRotate = new OctLoc[MAX_FACES][COORD];
	OctNextRowP3[][] slideNextRowP3 =
		new OctNextRowP3[MAX_SIDES][MAX_ORIENT];
	static final int[][] reverseP3 = {
		{IGNORE, 0, 0, 1, IGNORE, 1, 1, 1},
		{1, 1, IGNORE, 0, 0, 1, IGNORE, 1},
		{IGNORE, 1, 1, 1, IGNORE, 0, 0, 1},
		{0, 1, IGNORE, 1, 1, 1, IGNORE, 0}
	};
	static final int[][] rotOrientRowP3 = {
		/* current orient, rotation */
		{IGNORE, 11, 7, 2, IGNORE, 10, 5, 1},
		{11, 7, IGNORE, 5, 1, 2, IGNORE, 10},
		{IGNORE, 10, 5, 1, IGNORE, 11, 7, 2},
		{1, 2, IGNORE, 10, 11, 7, IGNORE, 5}
	};
	OctLoc[][] facetLoc = null;
	OctLoc[][] startLoc = null;
	OctLoc[] faceLoc = null;
	OctLoc[][] rowLoc = null;
	Point[][] triangleUnit = new Point[MAX_SIDES][4];
	int[][] triangleListX = new int[MAX_SIDES][4];
	int[][] triangleListY = new int[MAX_SIDES][4];
	Point[] letterList = new Point[MAX_SIDES];

	@Override
	protected String paramString() {
		return super.paramString() +
		",size=" + size +
		",period=" + period +
		",orient=" + orient +
		",sticky=" + sticky +
		",practice=" + practice +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + faceNames +
		",color0=" + faceColor[0] +
		",color1=" + faceColor[1] +
		",color2=" + faceColor[2] +
		",color3=" + faceColor[3] +
		",color4=" + faceColor[4] +
		",color5=" + faceColor[5] +
		",color6=" + faceColor[6] +
		",color7=" + faceColor[7] +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkPieces() {
		if (size < OctInterface.MIN_FACETS) {
			System.out.println(size +
				" number of pieces on a edge out of bounds, must be at least " +
				OctInterface.MIN_FACETS +
				", defaulting to " +
				OctInterface.DEFAULTFACETS);
			size = OctInterface.DEFAULTFACETS;
		}
		if (period < OctInterface.PERIOD3 ||
				period > OctInterface.BOTH) {
			System.out.println("Mode " + period +
				" is in error, use 3 for Period3, 4 for Period4, 5 for Both, defaulting to " +
				OctInterface.DEFAULT_MODE);
			period = OctInterface.DEFAULT_MODE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved() {
		for (int face = 0; face < MAX_FACES; face++) {
			int tempFace = facetLoc[face][0].face;
			int tempRotation = facetLoc[face][0].rotation;

			for (int position = 1; position < sizeSize;
					position++) {
				if (tempFace != /* MAX_SIDES * view + face */
						facetLoc[face][position].face ||
						(orient && tempRotation !=
						facetLoc[face][position].rotation))
					return false;
			}
		}
		return true;
	}

	/* This is fast for small i, a -1 is returned for negative i. */
	static int SQRT(int i) {
		int j = 0;

		while (j * j <= i)
			j++;
		return (j - 1);
	}

	static OctRTT toRTT(int position) {
		int row = SQRT(position);

		/* Passing row so there is no sqrt calculation again */
		return (new OctRTT(row,
			(position - row * row) / 2,
			(row * row + 2 * row - position) / 2));
	}

	static int toPosition(OctRTT rtt) {
		return (rtt.row * rtt.row + rtt.row + rtt.trbl - rtt.tlbr);
	}

	int length(int dir, int h) {
		if ((dir % 2 == 0) || dir > COORD)
			return (2 * h + 1);
		return (2 * (size - h) - 1);
	}

	static void fillPolygon(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(fillColor);
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor(lineColor);
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawOrientLine(Graphics g, int orientLine, int face, int side,
			int dx, int dy) {
		int x1 = 0, x2 = 0, y1 = 0, y2 = 0;
		int temp1 = facetLength + 1;
		int temp2 = facetLength / 2 + 1;
		int temp3 = orientLineLength / 3;

		/* clock positions */
		switch ((side == face) ? orientLine :
			((orientLine + MAX_ORIENT / 2) % MAX_ORIENT)) {
		case 0:
			x2 = x1 = dx;
			y1 = dy + temp1 - 2;
			y2 = y1 - orientLineLength;
			break;
		case 1:
			x1 = dx + temp2;
			y1 = dy + temp2;
			x2 = x1 + temp3;
			y2 = y1 - orientLineLength;
			break;
		case 2:
			x1 = dx - temp2;
			y1 = dy - temp2;
			x2 = x1 + orientLineLength;
			y2 = y1 - temp3;
			break;
		case 3:
			x1 = dx - temp1 + 2;
			x2 = x1 + orientLineLength;
			y2 = y1 = dy;
			break;
		case 4:
			x1 = dx - temp2;
			y1 = dy + temp2;
			x2 = x1 + orientLineLength;
			y2 = y1 + temp3;
			break;
		case 5:
			x1 = dx + temp2;
			y1 = dy - temp2;
			x2 = x1 + temp3;
			y2 = y1 + orientLineLength;
			break;
		case 6:
			x2 = x1 = dx;
			y1 = dy - temp1 + 2;
			y2 = y1 + orientLineLength;
			break;
		case 7:
			x1 = dx - temp2;
			y1 = dy - temp2;
			x2 = x1 - temp3;
			y2 = y1 + orientLineLength;
			break;
		case 8:
			x1 = dx + temp2;
			y1 = dy + temp2;
			x2 = x1 - orientLineLength;
			y2 = y1 + temp3;
			break;
		case 9:
			x1 = dx + temp1 - 2;
			x2 = x1 - orientLineLength;
			y2 = y1 = dy;
			break;
		case 10:
			x1 = dx + temp2;
			y1 = dy - temp2;
			x2 = x1 - orientLineLength;
			y2 = y1 - temp3;
			break;
		case 11:
			x1 = dx - temp2;
			y1 = dy + temp2;
			x2 = x1 - temp3;
			y2 = y1 - orientLineLength;
			break;
		default:
			System.out.println("drawOrientLine: orient " + orientLine);
		}
		g.drawLine(x1, y1, x2, y2);
	}

	void drawTriangle(int face, int position, int offsetPosition) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			int side, row = SQRT(position);
			int base = row * row;
			int h = position - base;
			int s = h % 2;
			int pos = row * (facetLength + delta);
			int rotate, view, faceOnView;

			view = (face / MAX_SIDES == UP) ? DOWN : UP;
			faceOnView = face - MAX_SIDES *
				((view == UP) ? DOWN : UP);
			rotate = (!vertical && view == DOWN) ?
				(faceOnView + MAX_SIDES / 2) % MAX_SIDES :
				faceOnView;
			Point start = new Point(0, 0);
			switch (rotate) {
			case 0:
				start.y = viewMiddle - delta - 2 - pos;
				start.x = viewMiddle + pos - 1 -
					h * (facetLength + delta);
				break;
			case 1:
				start.x = viewMiddle + delta + pos;
				start.y = viewMiddle + pos - 1 -
					h * (facetLength + delta);
				break;
			case 2:
				start.y = viewMiddle + delta + pos;
				start.x = viewMiddle - pos - 1 +
					h * (facetLength + delta);
				break;
			case 3:
				start.x = viewMiddle - delta - 2 - pos;
				start.y = viewMiddle - pos - 1 +
					h * (facetLength + delta);
				break;
			default:
				System.out.println("drawTriangle: rotate " +
					rotate);
			}
			if ((faceOnView & 1) == 1)
				side = ((faceOnView == 1) ? 1 - s : s) * 2 + 1;
			else /* faceOnView == (RIGHT / 2) ||
				faceOnView == (LEFT / 2) */
				side = ((faceOnView == 0) ? 1 - s : s) * 2;
			side = (!vertical && view == DOWN) ?
				(side + MAX_SIDES / 2) % MAX_SIDES : side;
			if (s == OPPOSITE)
				switch (side) {
				case 0:
					start.y -= facetLength;
					break;
				case 1:
					start.x += facetLength;
					break;
				case 2:
					start.y += facetLength;
					break;
				case 3:
					start.x -= facetLength;
					break;
				default:
					System.out.println("drawTriangle: side "
						+ side);
				}
			start.x += puzzleOffset.x;
			start.y += puzzleOffset.y;
			if (view == DOWN) {
				if (vertical)
					start.y += viewLength - delta - 1;
				else
					start.x += viewLength - delta - 1;
			}


			int l = facetLoc[face][position].face;
			Color fillColor, lineColor;

			if (offsetPosition == 0) {
				if (mono) {
					if (reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = faceColor[l];
					lineColor = borderColor;
				}
			} else {
				if (mono) {
					if (reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = faceColor[l];
					fillColor = borderColor;
				}
			}
			fillPolygon(g, fillColor, lineColor,
				start.x,
				start.y,
				triangleListX[side],
				triangleListY[side],
				3);
			if (mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = facetLength / 3;
				g.setFont(new Font("arial", Font.PLAIN,
					fontSize));

				buf = Character.toString(faceNames.charAt(l));
				g.drawString(buf,
					start.x + letterList[side].x,
					start.y + letterList[side].y);
			}
			if (orient) {
				g.setColor(lineColor);
				drawOrientLine(g,
					facetLoc[face][position].rotation,
					faceOnView, side, start.x, start.y);
			}
		} finally {
			g.dispose();
		}
	}

	void drawFace(int face) {
		for (int position = 0; position < sizeSize; position++)
			drawTriangle(face, position, 0);
	}

	void drawAllPieces() {
		for (int face = 0; face < MAX_FACES; face++)
			drawFace(face);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			Point start = new Point(1 + puzzleOffset.x,
				1 + puzzleOffset.y);
			Point length = new Point(viewLength - delta +
				puzzleOffset.x, viewLength - delta +
				puzzleOffset.y);
			if (show) {
				g.setColor(currentForeground);
				g.drawLine(start.x, start.y, length.x, start.y);
				g.drawLine(start.x, start.y, start.x, length.y);
				g.drawLine(start.x, length.y, length.x, start.y);
				g.drawLine(start.x, start.y, length.x, length.y);
				int longlength = 2 * viewLength - 2 * delta - 1;
				if (vertical) {
					longlength += puzzleOffset.y;
					g.drawLine(start.x, length.y,
						start.x, longlength);
					g.drawLine(length.x, length.y,
						length.x, longlength);
					g.drawLine(start.x, longlength,
						length.x, longlength);
					g.drawLine(start.x, longlength,
						length.x, length.y);
					g.drawLine(start.x, length.y,
						length.x, longlength);
					g.drawLine(length.x, start.y,
						length.x, length.y);
					g.setColor(foreground);
					g.drawLine(0, length.y,
						coreSize.x, length.y);
				} else {
					longlength += puzzleOffset.x;
					g.drawLine(length.x, start.y,
						longlength, start.y);
					g.drawLine(length.x, length.y,
						longlength, length.y);
					g.drawLine(longlength, start.y,
						longlength, length.y);
					g.drawLine(longlength, start.y,
						length.x, length.y);
					g.drawLine(length.x, start.y,
						longlength, length.y);
					g.drawLine(start.x, length.y,
						length.x, length.y);
					g.setColor(foreground);
					g.drawLine(length.x, 0,
						length.x, coreSize.y);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void moveNoPieces() {
		((OctFrame) frame).callback(
			OctInterface.ACTION_ILLEGAL);
	}

	void rotateFace(int view, int side, int direction) {
		int g, square = 0;
		int i = 0, j = 0, k = -1, i1, j1, k1, position;

		/* Read Face */
		for (g = 0; g < sizeSize; g++) {
			/* This is the old algorithm, its now more efficient
			  k = SQRT(g);
			  j = (g - k * k) / 2;
			  i = ((k + 1) * (k + 1) - g - 1) / 2; */
			if (square <= g) {
				k++;
				square = (k + 1) * (k + 1);
				j = -1;
				i = k;
			}
			if (((square - g) & 1) == 0)
				i--;
			else
				j++;
			if (direction == CW) {
				k1 = size - 1 - i;
				i1 = j;
				j1 = size - 1 - k;
			} else { /* (direction == CCW) */
				k1 = size - 1 - j;
				j1 = i;
				i1 = size - 1 - k;
			}
			position = k1 * k1 + 2 * j1 + ((j1 == k1 - i1) ? 0 : 1);
			faceLoc[position] = facetLoc[view * MAX_SIDES + side][g];
		}
		/* Write Face */
		square = 1;
		for (g = 0; g < sizeSize; g++) {
			facetLoc[view * MAX_SIDES + side][g] = faceLoc[g];
			facetLoc[view * MAX_SIDES + side][g].rotation =
				(direction == CW) ?
				(facetLoc[view * MAX_SIDES + side][g].rotation + 4) % MAX_ORIENT :
				(facetLoc[view * MAX_SIDES + side][g].rotation + 8) % MAX_ORIENT;
			drawTriangle(view * MAX_SIDES + side, g, 0);
			if (g == square * square - 1) {
				++square;
			}
		}
	}

	void readRTT(int view, int side, int dir, int h, int len, int orientLine) {
		int f, g, s;
		int base = h * h;

		if (((dir & 1) == 0) || dir > COORD)
			for (g = 0; g < len; g++)
				rowLoc[orientLine][g] =
					facetLoc[view * MAX_SIDES + side][base + g];
		else if (((dir / 2 + side) & 1) == 1) {
			f = -1;
			for (g = 0; g < len; g++) {
				s = g & 1;
				rowLoc[orientLine][g] =
					facetLoc[view * MAX_SIDES + side][base + f + 1 - s];
				if (s == SAME)
					f += g + 2 * (h + 1) + 1;
			}
		} else {
			base += 2 * h;
			f = 1;
			for (g = 0; g < len; g++) {
				s = g & 1;
				rowLoc[orientLine][g] =
					facetLoc[view * MAX_SIDES + side][base + f - 1 + s];
				if (s == SAME)
					f += g + 2 * h + 1;
			}
		}
	}

	void rotateRTT(int rotate, int len, int orientLine) {
		for (int g = 0; g < len; g++)
			rowLoc[orientLine][g].rotation =
				(rowLoc[orientLine][g].rotation + rotate) %
				MAX_ORIENT;
	}

	void reverseRTT(int len, int orientLine) {
		OctLoc temp;

		for (int g = 0; g < (len - 1) / 2; g++) {
			temp = rowLoc[orientLine][len - 1 - g];
			rowLoc[orientLine][len - 1 - g] = rowLoc[orientLine][g];
			rowLoc[orientLine][g] = temp;
		}
	}

	void writeRTT(int view, int side, int dir, int h, int len, int orientLine) {
		int f, g, s;
		int base = h * h;

		if (((dir & 1) == 0) || dir > COORD) { /* CW || CCW */
			for (g = 0; g < len; g++) {
				//s = g & 1;
				facetLoc[view * MAX_SIDES + side][base + g] =
					rowLoc[orientLine][g];
				drawTriangle(view * MAX_SIDES + side,
					base + g, 0);
			}
		} else if (((dir / 2 + side) & 1) == 1) {
			f = -1;
			for (g = 0; g < len; g++) {
				s = g & 1;
				facetLoc[view * MAX_SIDES + side][base + f + 1 - s] =
					rowLoc[orientLine][g];
				drawTriangle(view * MAX_SIDES + side,
					base + f + 1 - s, 0);
				if (s == SAME)
					f += g + 2 * (h + 1) + 1;
			}
		} else {
			base += 2 * h;
			f = 1;
			for (g = 0; g < len; g++) {
				s = g & 1;
				facetLoc[view * MAX_SIDES + side][base + f - 1 + s] =
					rowLoc[orientLine][g];
				drawTriangle(view * MAX_SIDES + side,
					base + f - 1 + s, 0);
				if (s == SAME)
					f += g + 2 * h + 1;
			}
		}
	}

	void movePieces(int face, OctRTT rttRotate, int direction, int style) {
		boolean bound;
		int view, side, orientLine, newView, rotate, g, h, len;
		int newSide, oldSide, newDirection, l = 0;
		int dir = direction;

		view = face / MAX_SIDES;
		side = face % MAX_SIDES;
		if (style == OctInterface.PERIOD3) {
			if (dir == CW || dir == CCW) {
				/* Remap to row movement */
				side = (side + 2) % MAX_SIDES;
				dir = ((side + dir) % MAX_SIDES) * 2;
				/*face = side + MAX_SIDES * ((view == UP) ? DOWN : UP);*/
				rttRotate.row = rttRotate.trbl = rttRotate.tlbr = 0;
			}
			if (((rttRotate.row == 0) && ((dir & 1) == 0)) ||
			    ((rttRotate.trbl == 0) && ((dir & 1) == 1) &&
			    (((side + dir / 2) & 1) == 0)) ||
			    ((rttRotate.tlbr == 0) && ((dir & 1) == 1) &&
			    (((side + dir / 2) & 1) == 1)))
				rotateFace(view,
					rowToRotate[side][dir].face,
					rowToRotate[side][dir].rotation);
			if ((rttRotate.row == size - 1 && ((dir & 1) == 0)) ||
			    (rttRotate.trbl == size - 1 && ((dir & 1) == 1) &&
			    (((side + dir / 2) & 1) == 0)) ||
			    (rttRotate.tlbr == size - 1 && ((dir & 1) == 1) &&
			     (((side + dir / 2) & 1) == 1)))
				rotateFace((view == UP) ? DOWN : UP,
					rowToRotate[MAX_SIDES + side][dir].face,
				  rowToRotate[MAX_SIDES + side][dir].rotation);
			if ((dir & 1) == 0)
				h = rttRotate.row;
			else if (((dir / 2 + side) & 1) == 1)
				h = rttRotate.tlbr;
			else /* (((dir / 2 + side) & 1) == 0) */
				h = rttRotate.trbl;
			if (sticky && (h == 1 || h == 2)) {
				l = 0;
				bound = true;
				h = 1;
			} else
				bound = false;
			newView = view;
			newSide = side;
			newDirection = dir;
			do {
				len = length(dir, h);
				rotate = rotOrientRowP3[side][dir];
				readRTT(view, side, dir, h, len, 0);
				if (reverseP3[side][dir] == 1)
					reverseRTT(len, 0);
				rotateRTT(rotate, len, 0);
				for (orientLine = 1; orientLine < 8; orientLine++) {
					if (slideNextRowP3[side][dir].viewChanged) {
						view = (view == UP) ? DOWN : UP;
						h = size - 1 - h;
					}
					oldSide = slideNextRowP3[side][dir].face;
					dir = slideNextRowP3[side][dir].direction;
					side = oldSide;
					len = length(dir, h);
					rotate = rotOrientRowP3[side][dir];
					if (orientLine < 6) {
						readRTT(view, side, dir, h, len, orientLine);
						if (reverseP3[side][dir] == 1)
							reverseRTT(len, orientLine);
						rotateRTT(rotate, len, orientLine);
					}
					if (orientLine >= 2)
						writeRTT(view, side, dir, h, len, orientLine - 2);
				}
				l++;
				h = 2;
				view = newView;
				side = newSide;
				dir = newDirection;
			} while (bound && l < 2);
		} else { /* style == OctInterface.PERIOD4 */
			if (dir > TL)
				h = rttRotate.row;
			else if (((dir / 2 + side) & 1) == 1)
				h = rttRotate.tlbr;
			else /* (((dir / 2 + side) & 1) == 0) */
				h = rttRotate.trbl;

			if (sticky && !((dir > TL && h == size - 1) ||
					(dir <= TL && h == 0))) {
				l = 0;
				h = (dir <= TL) ? 1 : 0;
				bound = true;
			} else
				bound = false;
			g = 0;
			do { /* In case this is on an edge */
			  len = length(dir, h);
			  if (g == 1) {
			    if (dir > TL) {
			      dir = (dir == CW) ? CCW : CW;
			      view = (view == UP) ? DOWN : UP;
			    } else
			      side = (side + 2) % MAX_SIDES;
			  }
			  readRTT(view, side, dir, h, len, 0);
			  for (orientLine = 1; orientLine <= 4; orientLine++) {
			    if (dir <= TL) {
			      if ((side - dir / 2 + COORD) %
				   MAX_SIDES < 2) {
				newView = (view == UP) ? DOWN : UP;
				newSide = ((side & 1) == 1) ?
				  side : (side + 2) % MAX_SIDES;
				newDirection = (((dir / 2) & 1) == 0) ?
				  ((dir + 6) % MAX_FACES) : ((dir + 2) % MAX_FACES);
				if ((side & 1) == 1)
				  rotate = (((newDirection - dir) / 2 + MAX_SIDES) %
				    MAX_SIDES == 1) ? 4 : MAX_ORIENT - 4;
				else
				  rotate = (((newDirection - dir) / 2 + MAX_SIDES) %
				    MAX_SIDES == 1) ? 2 : MAX_ORIENT - 2;
			      } else {	/* Next is on same view */
				newView = view;
				newSide = MAX_SIDES - side - 1;
				if (((dir / 2) & 1) == 1)
				  newSide = (newSide + 2) % MAX_SIDES;
				newDirection = dir;
				rotate = ((side - newSide + MAX_SIDES) % MAX_SIDES == 1) ?
				  1 : MAX_ORIENT - 1;
			      }
			    } else { /* dir == CW || dir == CCW */
			      newView = view;
			      newSide = (side + dir) % MAX_SIDES;
			      newDirection = dir;
			      rotate = 3 * newDirection;
			    }
			    if (orientLine != 4)
			      readRTT(newView, newSide, newDirection, h, len, orientLine);
			    rotateRTT(rotate, len, orientLine - 1);
			    if (dir <= TL)
			      reverseRTT(len, orientLine - 1);
			    writeRTT(newView, newSide, newDirection, h, len, orientLine - 1);
			    view = newView;
			    side = newSide;
			    dir = newDirection;
			  }
			  l++;
			  if (sticky &&
			      !((dir > TL && h == size - 1) ||
			      (dir <= TL && h == 0)))
			    h++;
			  else
			    g++;
			} while ((bound && l < size - 1) ||
			   (((dir > TL && h == size - 1) ||
			   (dir <= TL && h == 0)) && g < 2 && !bound));
		}
	}

	void moveControlCb(int face, int direction, int style) {
		OctRTT rttTemp = new OctRTT(-1, -1, -1);

		if (sticky) {
			if (style == OctInterface.PERIOD3)
				for (int i = 0; i < 3; i++) {
				  if (i == 2)
				    i++;
				  if (direction == CW || direction == CCW) {
				    if (face >= MAX_SIDES)
				      direction = (direction == CW) ? CCW : CW;
				      if (face == 5)
					face = 7;
				      else if (face == 7)
					face = 5;
				      face = (face + 2) % MAX_SIDES;
				      direction = ((face + direction) % MAX_SIDES) * 2;
				  }
				  rttTemp.row = i;
				  rttTemp.trbl = i;
				  rttTemp.tlbr = i;
				  movePieces(face, rttTemp, direction, style);
				  ((OctFrame) frame).callback(
				    OctInterface.ACTION_CONTROL);
				}
			else /* (style == OctInterface.PERIOD4) */
				for (int i = 0; i < 3; i++) {
				  if (i == 2) {
				    if (direction == CW || direction == CCW) {
				      int j = i - 2;
				      rttTemp.row = j;
				      rttTemp.trbl = j;
				      rttTemp.tlbr = j;
				      movePieces((1 - face / MAX_SIDES) *
					MAX_SIDES + (((face & 1) == 1) ?
					(face + MAX_SIDES / 2) % MAX_SIDES :
					face % MAX_SIDES), rttTemp,
					(direction == CW) ? CCW : CW, style);
				    } else {
				      int j = i + 1;
				      rttTemp.row = j;
				      rttTemp.trbl = j;
				      rttTemp.tlbr = j;
				      movePieces((1 - face / MAX_SIDES) *
					MAX_SIDES + (((face & 1) == 1) ?
					(face + MAX_SIDES / 2) % MAX_SIDES :
					face % MAX_SIDES), rttTemp,
					(((direction / 2) & 1) == 1) ?
					(direction + 2) % MAX_FACES :
					(direction + 6) % MAX_FACES, style);
				    }
				  } else {
				    int j;

				    if (i == 1)
				      j = i + 2;
				    else
				      j = i;
				    rttTemp.row = j;
				    rttTemp.trbl = j;
				    rttTemp.tlbr = j;
				    movePieces(face, rttTemp, direction, style);
				  }
				  ((OctFrame) frame).callback(
				    OctInterface.ACTION_CONTROL);
				}
		} else {
			if (style == OctInterface.PERIOD3)
				for (int i = 0; i < size; i++) {
				  if (direction == CW || direction == CCW) {
				    if (face >= MAX_SIDES)
				      direction = (direction == CW) ? CCW : CW;
				      if (face == 5)
					face = 7;
				      else if (face == 7)
					face = 5;
				      face = (face + 2) % MAX_SIDES;
				      direction = ((face + direction) % MAX_SIDES) * 2;
				  }
				  rttTemp.row = i;
				  rttTemp.trbl = i;
				  rttTemp.tlbr = i;
				  movePieces(face, rttTemp, direction, style);
				  ((OctFrame) frame).callback(
				    OctInterface.ACTION_CONTROL);
				}
			else /* (style == OctInterface.PERIOD4) */
				for (int i = 0; i < 2 * size - 1; i++) {
				  if (i < size) {
				    rttTemp.row = i;
				    rttTemp.trbl = i;
				    rttTemp.tlbr = i;
				    movePieces(face, rttTemp, direction, style);
				  } else {
				    if (direction == CW || direction == CCW) {
				      int j = i - size;
				      rttTemp.row = j;
				      rttTemp.trbl = j;
				      rttTemp.tlbr = j;
				      movePieces((1 - face / MAX_SIDES) *
					MAX_SIDES + (((face & 1) == 1) ?
					(face + MAX_SIDES / 2) % MAX_SIDES :
					face % MAX_SIDES), rttTemp,
					(direction == CW) ? CCW : CW, style);
				    } else {
				      int j = i - size + 1;
				      rttTemp.row = j;
				      rttTemp.trbl = j;
				      rttTemp.tlbr = j;
				      movePieces((1 - face / MAX_SIDES) *
					MAX_SIDES + (((face & 1) == 1) ?
					(face + MAX_SIDES / 2) % MAX_SIDES :
					face % MAX_SIDES), rttTemp,
					(((direction / 2) & 1) == 1) ?
					(direction + 2) % MAX_FACES :
					(direction + 6) % MAX_FACES, style);
				    }
				  }
				  ((OctFrame) frame).callback(
				    OctInterface.ACTION_CONTROL);
				}
		}
	}

	void movePuzzle(int face, int position, int direction, int style,
			boolean control) {
		if (control) {
			moveControlCb(face, direction, style);
		} else {
			movePieces(face, toRTT(position), direction, style);
			((OctFrame) frame).callback(
				OctInterface.ACTION_MOVED);
		}
		undo.setMove(direction, style, (control) ? 1 : 0,
			face, position);
		redo.flushMoves();
		if (((OctFrame) frame).getToggleSound())
			((OctFrame) frame).playMoveAudio();
	}

	void movePuzzleDelay(int face, int position, int direction, int style,
			boolean control, int fast) {
		movePuzzle(face, position, direction, style, control);
		try {
			Thread.sleep(64 * delay / fast);
		} catch (InterruptedException e) {
			// ignore
		}
	}

	void selectPieces() {
		if (!nextControl && !practice && checkSolved())
			moveNoPieces();

		else if (nextFace >= 0) {
			int count;

			if (period == OctInterface.BOTH) {
				currentStyle = (nextShift) ?
					OctInterface.PERIOD4 :
					OctInterface.PERIOD3;
			} else {
				if (nextControl && nextShift)
					currentStyle = (period ==
						OctInterface.PERIOD4) ?
						OctInterface.PERIOD3 :
						OctInterface.PERIOD4;
				else
					currentStyle = (period ==
						OctInterface.PERIOD3) ?
						OctInterface.PERIOD3 :
						OctInterface.PERIOD4;
			}
			// next sets currentDirection
			count = checkMoveDir(currentFace, toRTT(currentPosition),
				nextFace, toRTT(nextPosition), currentStyle);
			switch (count) {
			case 1:
				narrowSelection(currentStyle);
				movePuzzle(currentFace, currentPosition,
					currentDirection, currentStyle,
					nextControl);
				started = true;
				if (!nextControl && checkSolved()) {
					((OctFrame) frame).callback(
						OctInterface.ACTION_SOLVED);
				}
				break;
			case 2:
				/* count == 3 too annoying */
				((OctFrame) frame).callback(
					OctInterface.ACTION_AMBIGUOUS);
				break;
			case 0:
				moveNoPieces();
				break;
			default:
				break;
			}
		}
	}

	int checkMoveDir(int face1, OctRTT rtt1,
			int face2, OctRTT rtt2, int style) {
		// sets currentDirection
		int which = -1, count = 0;

		if (face1 == face2) {
			if (rtt1.row == rtt2.row) {
				which = 0;
				count++;
			}
			if (rtt1.trbl == rtt2.trbl) {
				which = 1;
				count++;
			}
			if (rtt1.tlbr == rtt2.tlbr) {
				which = 2;
				count++;
			}
			if (count == 1)
				switch (which) {
				case 0:	/* ROW */
					if (rtt2.trbl > rtt1.trbl)
						currentDirection =
							(2 * face1 + LEFT) %
							COORD;
					else
						currentDirection =
							(2 * face1 + RIGHT) %
							COORD;
					break;
				case 1:	/* TRBL */
					if (rtt2.row > rtt1.row)
						currentDirection =
							(2 * face1 + TR) %
							COORD;
					else
						currentDirection =
							(2 * face1 + BL) %
							COORD;
					break;
				case 2:	/* TLBR */
					if (rtt2.row > rtt1.row)
						currentDirection =
							(2 * face1 + TL) %
							COORD;
					else
						currentDirection =
							(2 * face1 + BR) %
							COORD;
					break;
				default:
					currentDirection = -1;
					return 0;
			}
			if (!vertical && face1 >= MAX_SIDES &&
					currentDirection > TL)
				currentDirection =
					(currentDirection + MAX_SIDES) %
					MAX_FACES;
		} else {
			if (style == 3) {
				if ((face1 == 0 && face2 == 1) ||
						(face1 == 2 && face2 == 3) ||
						(face1 == 4 && face2 == 5) ||
						(face1 == 6 && face2 == 7)) {
					if (rtt1.row == rtt2.trbl) {
						which = 0;
						count++;
					}
					if (rtt1.tlbr == rtt2.row) {
						which = 2;
						count++;
					}
				} else if ((face1 == 0 && face2 == 3) ||
						(face1 == 2 && face2 == 1) ||
						(face1 == 4 && face2 == 7) ||
						(face1 == 6 && face2 == 5)) {
					if (rtt1.row == rtt2.tlbr) {
						which = 0;
						count++;
					}
					if (rtt1.trbl == rtt2.row) {
						which = 1;
						count++;
					}
				} else if ((face1 == 0 && face2 == 6) ||
						(face1 == 2 && face2 == 4) ||
						(face1 == 4 && face2 == 2) ||
						(face1 == 6 && face2 == 0)) {
					if (rtt1.trbl == size - 1 -
							rtt2.trbl) {
						which = 1;
						count++;
					}
					if (rtt1.tlbr == size - 1 -
							rtt2.tlbr) {
						which = 2;
						count++;
					}
				} else if ((face1 == 1 && face2 == 0) ||
						(face1 == 3 && face2 == 2) ||
						(face1 == 5 && face2 == 4) ||
						(face1 == 7 && face2 == 6)) {
					if (rtt1.row == rtt2.tlbr) {
						which = 0;
						count++;
					}
					if (rtt1.trbl == rtt2.row) {
						which = 2;
						count++;
					}
				} else if ((face1 == 1 && face2 == 2) ||
						(face1 == 3 && face2 == 0) ||
						(face1 == 5 && face2 == 6) ||
						(face1 == 7 && face2 == 4)) {
					if (rtt1.row == rtt2.trbl) {
						which = 0;
						count++;
					}
					if (rtt1.tlbr == rtt2.row) {
						which = 1;
						count++;
					}
				} else if ((face1 == 1 && face2 == 5) ||
						(face1 == 3 && face2 == 7) ||
						(face1 == 5 && face2 == 1) ||
						(face1 == 7 && face2 == 3)) {
					if (rtt1.tlbr == size - 1 -
							rtt2.tlbr) {
						which = 1;
						count++;
					}
					if (rtt1.trbl == size - 1 -
							rtt2.trbl) {
						which = 2;
						count++;
					}
				/* P3 */
				} else if ((face1 == 0 && face2 == 2) ||
						(face1 == 2 && face2 == 0) ||
						(face1 == 4 && face2 == 6) ||
						(face1 == 6 && face2 == 4)) {
					if (rtt1.trbl == rtt2.tlbr) {
						which = 1;
						count++;
					}
					if (rtt1.tlbr == rtt2.trbl) {
						which = 2;
						count++;
					}
				} else if ((face1 == 0 && face2 == 5) ||
						(face1 == 2 && face2 == 7) ||
						(face1 == 4 && face2 == 1) ||
						(face1 == 6 && face2 == 3)) {
					if (rtt1.row == size - 1 -
							rtt2.trbl) {
						which = 0;
						count++;
					}
					if (rtt1.trbl == size - 1 -
							rtt2.row) {
						which = 1;
						count++;
					}
				} else if ((face1 == 0 && face2 == 7) ||
						(face1 == 2 && face2 == 5) ||
						(face1 == 4 && face2 == 3) ||
						(face1 == 6 && face2 == 1)) {
					if (rtt1.row == size - 1 -
							rtt2.tlbr) {
						which = 0;
						count++;
					}
					if (rtt1.tlbr == size - 1 -
							rtt2.row) {
						which = 2;
						count++;
					}
				} else if ((face1 == 1 && face2 == 3) ||
						(face1 == 3 && face2 == 1) ||
						(face1 == 5 && face2 == 7) ||
						(face1 == 7 && face2 == 5)) {
					if (rtt1.tlbr == rtt2.trbl) {
						which = 1;
						count++;
					}
					if (rtt1.trbl == rtt2.tlbr) {
						which = 2;
						count++;
					}
				} else if ((face1 == 1 && face2 == 6) ||
						(face1 == 3 && face2 == 4) ||
						(face1 == 5 && face2 == 2) ||
						(face1 == 7 && face2 == 0)) {
					if (rtt1.row == size - 1 -
							rtt2.tlbr) {
						which = 0;
						count++;
					}
					if (rtt1.tlbr == size - 1 -
							rtt2.row) {
						which = 1;
						count++;
					}
				} else if ((face1 == 1 && face2 == 4) ||
						(face1 == 3 && face2 == 6) ||
						(face1 == 5 && face2 == 0) ||
						(face1 == 7 && face2 == 2)) {
					if (rtt1.row == size - 1 -
							rtt2.trbl) {
						which = 0;
						count++;
					}
					if (rtt1.trbl == size - 1 -
							rtt2.row) {
						which = 2;
						count++;
					}
				}
			} else /* if (style == 4) */ {
				if ((face1 == 0 && face2 == 1) ||
						(face1 == 2 && face2 == 3) ||
						(face1 == 4 && face2 == 5) ||
						(face1 == 6 && face2 == 7)) {
					if (rtt1.row == rtt2.row) {
						which = 0;
						count++;
					}
					if (rtt1.tlbr == rtt2.trbl) {
						which = 2;
						count++;
					}
				} else if ((face1 == 0 && face2 == 3) ||
						(face1 == 2 && face2 == 1) ||
						(face1 == 4 && face2 == 7) ||
						(face1 == 6 && face2 == 5)) {
					if (rtt1.row == rtt2.row) {
						which = 0;
						count++;
					}
					if (rtt1.trbl == rtt2.tlbr) {
						which = 1;
						count++;
					}
				} else if ((face1 == 0 && face2 == 6) ||
						(face1 == 2 && face2 == 4) ||
						(face1 == 4 && face2 == 2) ||
						(face1 == 6 && face2 == 0)) {
					if (rtt1.trbl == rtt2.tlbr) {
						which = 1;
						count++;
					}
					if (rtt1.tlbr == rtt2.trbl) {
						which = 2;
						count++;
					}
				} else if ((face1 == 1 && face2 == 0) ||
						(face1 == 3 && face2 == 2) ||
						(face1 == 5 && face2 == 4) ||
						(face1 == 7 && face2 == 6)) {
					if (rtt1.row == rtt2.row) {
						which = 0;
						count++;
					}
					if (rtt1.trbl == rtt2.tlbr) {
						which = 2;
						count++;
					}
				} else if ((face1 == 1 && face2 == 2) ||
						(face1 == 3 && face2 == 0) ||
						(face1 == 5 && face2 == 6) ||
						(face1 == 7 && face2 == 4)) {
					if (rtt1.row == rtt2.row) {
						which = 0;
						count++;
					}
					if (rtt1.tlbr == rtt2.trbl) {
						which = 1;
						count++;
					}
				} else if ((face1 == 1 && face2 == 5) ||
						(face1 == 3 && face2 == 7) ||
						(face1 == 5 && face2 == 1) ||
						(face1 == 7 && face2 == 3)) {
					if (rtt1.tlbr == rtt2.trbl) {
						which = 1;
						count++;
					}
					if (rtt1.trbl == rtt2.tlbr) {
						which = 2;
						count++;
					}
				}
			}
			if (count == 1) {
				switch (which) {
				case 0: /* ROW */
					if (directionNextFace[TOP][face1] ==
							face2) {
						currentDirection = TOP;
					} else if (directionNextFace[RIGHT][face1] ==
							face2) {
						currentDirection = RIGHT;
					} else if (directionNextFace[BOTTOM][face1] ==
							face2) {
						currentDirection = BOTTOM;
					} else if (directionNextFace[LEFT][face1] ==
							face2) {
						currentDirection = LEFT;
					}
					if (style == 3) {
					if (directionNextFaceP3[TOP][face1] ==
							face2) {
						currentDirection = TOP;
					} else if (directionNextFaceP3[RIGHT][face1] ==
							face2) {
						currentDirection = RIGHT;
					} else if (directionNextFaceP3[BOTTOM][face1] ==
							face2) {
						currentDirection = BOTTOM;
					} else if (directionNextFaceP3[LEFT][face1] ==
							face2) {
						currentDirection = LEFT;
					}
					}
					break;
				case 1: /* TRBL */
					if (directionNextFace[TR][face1] ==
							face2) {
						currentDirection = TR;
					} else if (directionNextFace[BL][face1] ==
							face2) {
						currentDirection = BL;
					}
					if (style == 3) {
					if (directionNextFaceP3[TR][face1] ==
							face2) {
						currentDirection = TR;
					} else if (directionNextFaceP3[BL][face1] ==
							face2) {
						currentDirection = BL;
					}
					}
					break;
				case 2: /* TLBR */
					if (directionNextFace[TL][face1] ==
							face2) {
						currentDirection = TL;
					} else if (directionNextFace[BR][face1] ==
							face2) {
						currentDirection = BR;
					}
					if (style == 3) {
					if (directionNextFaceP3[TL][face1] ==
							face2) {
						currentDirection = TL;
					} else if (directionNextFaceP3[BR][face1] ==
							face2) {
						currentDirection = BR;
					}
					}
					break;
				default:
					count = 0;
				}
			}
			if (count >= 3)
				/* Very ambiguous but lets not ignore due to face change */
				count = 2;
		}
		return count;
	}

	void narrowSelection(int style) {
		// sets currentFace, rtt, currentDirection
		int side = currentFace % MAX_SIDES;

		if (((currentDirection & 1) == 0) &&
				((currentDirection / 2 + side) & 1) == 0)
			currentFace = -1;
		if (style == OctInterface.PERIOD4 &&
				(currentDirection & 1) == 0) {
			if (currentDirection == (2 * side + 2) %
				MAX_FACES)
				currentDirection = CW;
			else /* currentDirection == (2 * side + 6) %
				MAX_FACES */
				currentDirection = CCW;
		}
		/*
		if (style == OctInterface.PERIOD4) {
			if (currentDirection > TL)
				h = rtt.row;
			else if (((currentDirection / 2) & 1) == 0)
				h = rtt.trbl;
			else *//* ((currrentDirection / 2) & 1) == 1 *//*
				h = rtt.tlbr;
		} */
	}

	int positionToPiece(int positionX, int positionY) { // sets rtt
		int view, face;
		int x = positionX, y = positionY;

		x -= puzzleOffset.x;
		y -= puzzleOffset.y;
		if (vertical && y > viewLength - 1) {
			y -= (viewLength - 1);
			view = DOWN;
		} else if (!vertical && x > viewLength - 1) {
			x -= (viewLength - 1);
			view = DOWN;
		} else
			view = UP;
		if (x <= 0 || y <= 0 ||
				x >= faceLength + delta ||
				y >= faceLength + delta)
			return -1;
		else if (x + y > faceLength) {
			if (x > y)
				face = 1;
			else if (x < y)
				face = 2;
			else
				return -1;
		} else {
			if (x > y)
				face = 0;
			else if (x < y)
				face = 3;
			else
				return -1;
		}
		rtt.row = 0;
		while ((x <= ((size - (rtt.row + 1)) *
				(facetLength + delta)) ||
			 x >= (viewMiddle + (rtt.row + 1) *
				(facetLength + delta) + 1) ||
			 y <= ((size - (rtt.row + 1)) *
				(facetLength + delta)) ||
			 y >= (viewMiddle + (rtt.row + 1) *
				(facetLength + delta) + 1)) &&
				rtt.row < size)
			rtt.row++;
		rtt.trbl = 0;
		while (((x + y) / 2) <= ((size - (rtt.trbl + 1)) *
				(facetLength + delta) + 2 * delta) ||
			((x + y) / 2) >= (viewMiddle + (rtt.trbl + 1) *
				(facetLength + delta)))
			rtt.trbl++;
		rtt.tlbr = 0;
		while (x <= (y - 2 * (rtt.tlbr + 1) *
				(facetLength + delta) - 2) ||
			y <= (x - 2 * (rtt.tlbr + 1) *
				(facetLength + delta) - 2))
			rtt.tlbr++;
		if (!vertical && view == DOWN)
			face = (face + MAX_SIDES / 2) % MAX_SIDES;
		face += MAX_SIDES * ((view == UP) ? DOWN : UP);
		if ((face & 1) == 1) {
			view = rtt.tlbr;
			rtt.tlbr = rtt.trbl;
			rtt.trbl = view;
		}
		return face;
	}

	boolean positionPieces(int x, int y, int style, int direction) {
		// sets currentFace, rtt, currentDirection
		if ((currentFace = positionToPiece(x, y)) == -1)
			return false;
		currentDirection = direction;
		if (!vertical && currentFace >= MAX_SIDES &&
				currentDirection < MAX_FACES)
			currentDirection = (currentDirection + MAX_SIDES) %
				MAX_FACES;
		narrowSelection(style); // sets currentFace, rtt, currentDirection
		if (currentFace == -1) {
			moveNoPieces();
			return false;
		}
		return true;
	}

	void movePuzzleInput(int x, int y, int direction, boolean shift,
			boolean control) {
		int style;

		if (period == OctInterface.BOTH) {
			style = (shift) ? OctInterface.PERIOD4 :
				OctInterface.PERIOD3;
		} else {
			if (control && shift)
				style = (period == OctInterface.PERIOD4) ?
					OctInterface.PERIOD3 :
					OctInterface.PERIOD4;
			else
				style = (period == OctInterface.PERIOD3) ?
					OctInterface.PERIOD3 :
					OctInterface.PERIOD4;
		}
		if (!practice && !control && checkSolved()) {
			moveNoPieces();
			return;
		}
		if (!positionPieces(x, y, style, direction)) // sets currentFace, rtt, currentDirection
			return;
		movePuzzle(currentFace, toPosition(rtt), currentDirection, style, control);
		if (!control && checkSolved()) {
			((OctFrame) frame).callback(
				OctInterface.ACTION_SOLVED);
		}
	}

	void resetPieces() {
		solve.setFlag(false);
		sizeSize = size * size;
		facetLoc = new OctLoc[MAX_FACES][sizeSize];
		startLoc = new OctLoc[MAX_FACES][sizeSize];
		rowLoc = new OctLoc[MAX_ORIENT / 2][2 * size - 1];
		faceLoc = new OctLoc[sizeSize];
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeSize; position++) {
				startLoc[face][position] = new OctLoc(
					face, 3 * face % MAX_ORIENT);
				facetLoc[face][position] = new OctLoc(
					face, 3 * face % MAX_ORIENT);
			}
		for (int orientLine = 0; orientLine < MAX_ORIENT / 2; orientLine++)
			for (int i = 0; i < 2 * size - 1; i++)
				rowLoc[orientLine][i] = new OctLoc(0, 0);
		for (int i = 0; i < sizeSize; i++)
			faceLoc[i] = new OctLoc(0, 0);
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentFace = -1;
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void resizePieces() {
		facetLength = faceLength / (2 * size) - delta - 1;
		for (int i = 0; i <= 3; i++)
			for (int j = 0; j < MAX_SIDES; j++) {
				triangleListX[j][i] = triangleUnit[j][i].x *
					facetLength + ((i > 0) ?
					triangleListX[j][i - 1] : 0);
				triangleListY[j][i] = triangleUnit[j][i].y *
					facetLength + ((i > 0) ?
					triangleListY[j][i - 1] : 0);
			}
		letterList[TOP / 2] = new Point(-facetLength / 12 - 1,
			8 * facetLength / 12);
		letterList[RIGHT / 2] = new Point(-7 * facetLength / 12 - 2,
			1 * facetLength / 12 + 2);
		letterList[BOTTOM / 2] = new Point(-facetLength / 12 - 1,
			-6 * facetLength / 12 + 3);
		letterList[LEFT / 2] = new Point(5 * facetLength / 12,
			1 * facetLength / 12 + 2);
		orientLineLength = facetLength / 4;
	}

	void practicePieces() {
		((OctFrame) frame).callback(
			OctInterface.ACTION_PRACTICE);
	}

	void randomizePieces() {
		if (practice)
			practicePieces();
		((OctFrame) frame).callback(
			OctInterface.ACTION_RESET);
		randomPaints = java.lang.Math.min(sizeSize * 3,
			MAX_ITERATIONS) + generator.nextInt(2);
		if (sticky)
			randomPaints /= 3;
		repaint();
	}

	void randomizingPieces() {
		int position, randomDirection, style;

		int face = generator.nextInt(MAX_FACES);
		if (period == OctInterface.BOTH)
			style = generator.nextInt(OctInterface.MAX_MODES - 1) +
				OctInterface.PERIOD3;
		else
			style = period;
		if (sticky) {
			if (style == OctInterface.PERIOD3) {
				position = 6;
				randomDirection = generator.nextInt(6);
				if (randomDirection >= MAX_SIDES) {
					if (randomDirection == 4)
						randomDirection = CW;
					else if (randomDirection == 5)
						randomDirection = CCW;
				} else
					randomDirection = randomDirection * 2 + 1;
			} else { /* style == OctInterface.PERIOD4 */
				if (generator.nextInt(2) == 1) /* a point */
					position = 9;
				else /* the center */
					position = 6;
				randomDirection = generator.nextInt(6);
				switch (randomDirection) {
				case 4:
					randomDirection = CW;
					break;
				case 5:
					randomDirection = CCW;
					break;
				default:
					randomDirection = randomDirection * 2 + 1;
					position = 0;
					break;
				}
			}
		} else { /* (!sticky) */
			if (style == OctInterface.PERIOD3) {
				randomDirection = generator.nextInt(MAX_ORIENT / 2);
				if (randomDirection >= MAX_SIDES) {
					if (randomDirection == 4) {
						if ((face & 1) == 1)
							randomDirection = BOTTOM;
						else
							randomDirection = RIGHT;
					} else if (randomDirection == 5) {
						if ((face & 1) == 1)
							randomDirection = TOP;
						else
							randomDirection = LEFT;
					}
				} else {
					randomDirection = randomDirection * 2 + 1;
				}
			} else {	/* style == OctInterface.PERIOD4 */
				randomDirection = generator.nextInt(MAX_ORIENT / 2);
				switch (randomDirection) {
				case 4:
					randomDirection = CW;
					break;
				case 5:
					randomDirection = CCW;
					break;
				default:
					randomDirection = randomDirection * 2 + 1;
					break;
				}
			}
			position = generator.nextInt(sizeSize);
		}
		movePuzzle(face, position, randomDirection, style, false);
		randomPaints--;
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			// ignore
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((OctFrame) frame).callback(
				OctInterface.ACTION_RANDOMIZE);
			started = false;
			currentFace = -1;
			if (checkSolved())
				((OctFrame) frame).callback(
					OctInterface.ACTION_SOLVED);
		}
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
	}

	public boolean getCheat() {
		return cheat;
	}

	boolean setPuzzleSize(int newValue) {
		if (size != newValue &&
				newValue >= OctInterface.MIN_FACETS) {
			size = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePeriod(int newValue) {
		if (period != newValue &&
				newValue >= OctInterface.PERIOD3) {
			period = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleOrient(boolean newValue) {
		if (orient != newValue) {
			orient = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleSticky(boolean newValue) {
		if (sticky != newValue) {
			sticky = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePractice(boolean newValue) {
		if (practice != newValue) {
			practice = newValue;
			resetPieces();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((OctFrame) frame).callback(OctInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((OctFrame) frame).callback(OctInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((OctFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		int tempLength;

		coreSize = new Point(getSize().width, getSize().height);
		delta = 4;
		vertical = (getSize().height >= getSize().width);
		if (vertical)
			tempLength = java.lang.Math.min(getSize().height / 2,
				getSize().width);
		else
			tempLength = java.lang.Math.min(getSize().height,
				getSize().width / 2);
		facetLength = java.lang.Math.max((tempLength - delta + 1) /
			size, 0);
		faceLength = size * facetLength;
		viewLength = faceLength + delta + 3;
		viewMiddle = viewLength / 2;
		if (vertical) {
			puzzleSize = new Point(viewLength - 1,
				2 * viewLength - delta - 2);
		} else {
			puzzleSize = new Point(2 * viewLength - delta - 2,
				viewLength - 1);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		checkPieces();
		undo = new OctStack();
		redo = new OctStack();
		solve = new OctSolve(frame, this);
		resetPieces();
		generator = new Random(System.nanoTime());
		rowToRotate[0][0] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[0][1] = new OctLoc(1, CW);
		rowToRotate[0][2] = new OctLoc(2, CW);
		rowToRotate[0][3] = new OctLoc(3, CW);
		rowToRotate[0][4] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[0][5] = new OctLoc(1, CCW);
		rowToRotate[0][6] = new OctLoc(2, CCW);
		rowToRotate[0][7] = new OctLoc(3, CCW);

		rowToRotate[1][0] = new OctLoc(3, CCW);
		rowToRotate[1][1] = new OctLoc(0, CCW);
		rowToRotate[1][2] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[1][3] = new OctLoc(2, CW);
		rowToRotate[1][4] = new OctLoc(3, CW);
		rowToRotate[1][5] = new OctLoc(0, CW);
		rowToRotate[1][6] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[1][7] = new OctLoc(2, CCW);

		rowToRotate[2][0] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[2][1] = new OctLoc(3, CCW);
		rowToRotate[2][2] = new OctLoc(0, CCW);
		rowToRotate[2][3] = new OctLoc(1, CCW);
		rowToRotate[2][4] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[2][5] = new OctLoc(3, CW);
		rowToRotate[2][6] = new OctLoc(0, CW);
		rowToRotate[2][7] = new OctLoc(1, CW);

		rowToRotate[3][0] = new OctLoc(1, CW);
		rowToRotate[3][1] = new OctLoc(2, CW);
		rowToRotate[3][2] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[3][3] = new OctLoc(0, CCW);
		rowToRotate[3][4] = new OctLoc(1, CCW);
		rowToRotate[3][5] = new OctLoc(2, CCW);
		rowToRotate[3][6] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[3][7] = new OctLoc(0, CW);

		rowToRotate[4][0] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[4][1] = new OctLoc(3, CCW);
		rowToRotate[4][2] = new OctLoc(2, CCW);
		rowToRotate[4][3] = new OctLoc(1, CCW);
		rowToRotate[4][4] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[4][5] = new OctLoc(3, CW);
		rowToRotate[4][6] = new OctLoc(2, CW);
		rowToRotate[4][7] = new OctLoc(1, CW);

		rowToRotate[5][0] = new OctLoc(1, CW);
		rowToRotate[5][1] = new OctLoc(0, CW);
		rowToRotate[5][2] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[5][3] = new OctLoc(2, CCW);
		rowToRotate[5][4] = new OctLoc(1, CCW);
		rowToRotate[5][5] = new OctLoc(0, CCW);
		rowToRotate[5][6] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[5][7] = new OctLoc(2, CW);

		rowToRotate[6][0] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[6][1] = new OctLoc(1, CW);
		rowToRotate[6][2] = new OctLoc(0, CW);
		rowToRotate[6][3] = new OctLoc(3, CW);
		rowToRotate[6][4] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[6][5] = new OctLoc(1, CCW);
		rowToRotate[6][6] = new OctLoc(0, CCW);
		rowToRotate[6][7] = new OctLoc(3, CCW);

		rowToRotate[7][0] = new OctLoc(3, CCW);
		rowToRotate[7][1] = new OctLoc(2, CCW);
		rowToRotate[7][2] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[7][3] = new OctLoc(0, CW);
		rowToRotate[7][4] = new OctLoc(3, CW);
		rowToRotate[7][5] = new OctLoc(2, CW);
		rowToRotate[7][6] = new OctLoc(IGNORE, IGNORE);
		rowToRotate[7][7] = new OctLoc(0, CCW);

		slideNextRowP3[0][0] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[0][1] =
			new OctNextRowP3(true, 2, TR);
		slideNextRowP3[0][2] =
			new OctNextRowP3(false, 1, BR);
		slideNextRowP3[0][3] =
			new OctNextRowP3(false, 1, BOTTOM);
		slideNextRowP3[0][4] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[0][5] =
			new OctNextRowP3(false, 3, BOTTOM);
		slideNextRowP3[0][6] =
			new OctNextRowP3(false, 3, BL);
		slideNextRowP3[0][7] =
			new OctNextRowP3(true, 2, TL);
		slideNextRowP3[1][0] =
			new OctNextRowP3(false, 0, TL);
		slideNextRowP3[1][1] =
			new OctNextRowP3(true, 1, BL);
		slideNextRowP3[1][2] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[1][3] =
			new OctNextRowP3(true, 1, TL);
		slideNextRowP3[1][4] =
			new OctNextRowP3(false, 2, BL);
		slideNextRowP3[1][5] =
			new OctNextRowP3(false, 2, LEFT);
		slideNextRowP3[1][6] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[1][7] =
			new OctNextRowP3(false, 0, LEFT);
		slideNextRowP3[2][0] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[2][1] =
			new OctNextRowP3(false, 1, TOP);
		slideNextRowP3[2][2] =
			new OctNextRowP3(false, 1, TR);
		slideNextRowP3[2][3] =
			new OctNextRowP3(true, 0, BR);
		slideNextRowP3[2][4] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[2][5] =
			new OctNextRowP3(true, 0, BL);
		slideNextRowP3[2][6] =
			new OctNextRowP3(false, 3, TL);
		slideNextRowP3[2][7] =
			new OctNextRowP3(false, 3, TOP);
		slideNextRowP3[3][0] =
			new OctNextRowP3(false, 0, TR);
		slideNextRowP3[3][1] =
			new OctNextRowP3(false, 0, RIGHT);
		slideNextRowP3[3][2] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[3][3] =
			new OctNextRowP3(false, 2, RIGHT);
		slideNextRowP3[3][4] =
			new OctNextRowP3(false, 2, BR);
		slideNextRowP3[3][5] =
			new OctNextRowP3(true, 3, TR);
		slideNextRowP3[3][6] =
			new OctNextRowP3(false, IGNORE, IGNORE);
		slideNextRowP3[3][7] =
			new OctNextRowP3(true, 3, BR);
		triangleUnit[0][0] = new Point(0, 0);
		triangleUnit[0][1] = new Point(1, 1);
		triangleUnit[0][2] = new Point(-2, 0);
		triangleUnit[0][3] = new Point(1, -1);
		triangleUnit[1][0] = new Point(0, 0);
		triangleUnit[1][1] = new Point(-1, 1);
		triangleUnit[1][2] = new Point(0, -2);
		triangleUnit[1][3] = new Point(1, 1);
		triangleUnit[2][0] = new Point(0, 0);
		triangleUnit[2][1] = new Point(-1, -1);
		triangleUnit[2][2] = new Point(2, 0);
		triangleUnit[2][3] = new Point(-1, 1);
		triangleUnit[3][0] = new Point(0, 0);
		triangleUnit[3][1] = new Point(1, -1);
		triangleUnit[3][2] = new Point(0, 2);
		triangleUnit[3][3] = new Point(-1, -1);
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		drawAllPieces();
		solve.init();
	}

	void selectPuzzle(boolean control, int x, int y) {
		if ((currentFace = positionToPiece(x, y)) != -1) { //sets rtt
			currentPosition = toPosition(rtt);
			if (currentPosition >= sizeSize) {
				currentFace = -1;
				return;
			}
			if (control || practice || !checkSolved()) {
				selectPaint = true;
				repaint();
				return;
			}
			currentFace = -1;
		}
		currentFace = -1;
	}

	void releasePuzzle(boolean shift, boolean control, int x, int y) {
		if (currentFace == -1)
			return;
		nextShift = shift;
		nextControl = control;
		nextFace = positionToPiece(x, y); // sets rtt
		nextPosition = toPosition(rtt);
		releasePaint = true;
		repaint();
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizePieces();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle() {
		this.stringSave = ((OctFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read found " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, sizeValue, periodValue, moves;
			boolean orientValue, stickyValue, practiceValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			sizeValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (sizeValue >= OctInterface.MIN_FACETS) {
				for (i = this.size; i < sizeValue; i++) {
					((OctFrame) frame).callback(
						OctInterface.ACTION_INCREMENT);
				}
				for (i = this.size; i > sizeValue; i--) {
					((OctFrame) frame).callback(
						OctInterface.ACTION_DECREMENT);
				}
			} else {
				showMessage(source +
					" corrupted: size " + sizeValue +
					" should be at least " +
					OctInterface.MIN_FACETS);
				return;
			}
			lineData = token[++line];
			periodValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (periodValue >= OctInterface.PERIOD3 &&
					periodValue <= OctInterface.BOTH) {
				if (periodValue != this.period) {
					int reason = periodValue -
						OctInterface.PERIOD3 +
						OctInterface.ACTION_PERIOD3;

					((OctFrame) frame).callback(reason);
				}
			} else {
				showMessage(source +
					" corrupted: period " + periodValue +
					" should be even and between " +
					OctInterface.PERIOD3 +
					" and " + OctInterface.BOTH +
					" inclusive");
				return;
			}
			lineData = token[++line];
			stickyValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.sticky != stickyValue)
				((OctFrame) frame).callback(
					OctInterface.ACTION_STICKY);
			lineData = token[++line];
			orientValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orient != orientValue)
				((OctFrame) frame).callback(
					OctInterface.ACTION_ORIENTIZE);
			lineData = token[++line];
			practiceValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practice != practiceValue)
				((OctFrame) frame).callback(
					OctInterface.ACTION_PRACTICE);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((OctFrame) frame).callback(
				OctInterface.ACTION_RESTORE);
			setStartPosition();
			allPiecesPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": size " + sizeValue +
				", period " + periodValue + ", sticky " + stickyValue +
				", orient " + orientValue +
				", practice " + practiceValue + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("size").append(symbol).append(space)
			.append(size).append(newline);
		sb.append("period").append(symbol).append(space)
			.append(period).append(newline);
		sb.append("sticky").append(symbol).append(space)
			.append((sticky) ? 1 : 0).append(newline);
		sb.append("orient").append(symbol).append(space)
			.append((orient) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
			.append((practice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((OctFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Can not write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentStyle = undo.readMoveStyle();
		currentControl = (undo.readMoveControl() != 0);
		currentFace = undo.readMoveFace();
		currentPosition = undo.readMovePosition();
		redo.setMove(currentDirection, currentStyle, (currentControl) ?
			1 : 0, currentFace, currentPosition);
		// move in opposite direction
		currentDirection = (currentDirection < COORD) ?
			(currentDirection + MAX_SIDES) % MAX_FACES :
			3 * COORD - currentDirection;
		if (currentControl) {
			stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			stackPaint = true;
			repaint();
			started = true;
			((OctFrame) frame).callback(
				OctInterface.ACTION_UNDO);
		} /* else currentDirection = currentFace = -1; */
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentStyle = redo.readMoveStyle();
		currentControl = (redo.readMoveControl() != 0);
		currentFace = redo.readMoveFace();
		currentPosition = redo.readMovePosition();
		undo.setMove(currentDirection, currentStyle, (currentControl) ?
			1 : 0, currentFace, currentPosition);
		if (currentControl) {
			stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			stackPaint = true;
			repaint();
			started = true;
			((OctFrame) frame).callback(
				OctInterface.ACTION_REDO);
		} /* else currentDirection = currentFace = -1; */
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetPieces();
		drawAllPieces();
		allPiecesPaint = true;
		repaint();
	}

	void solvePuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0)
			return;
		if (((size <= 3 || sticky) && period == OctInterface.PERIOD3) ||
				(size <= 4 && period == OctInterface.PERIOD4)) {
			cheat = true;
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solvePieces();
		} else {
			showMessage("Auto-solver: sorry, only implemented for"
				+ NL + "size < 4 (or sticky mode set) where period = 3 or"
				+ NL + "size < 5 where  period = 4."
				+ NL);
		}
	}

	void incrementPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_INCREMENT);
	}

	void decrementPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		if (size <= OctInterface.MIN_FACETS)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_DECREMENT);
	}

	void period3ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_PERIOD3);
	}

	void period4ModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_PERIOD4);
	}

	void bothModePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_BOTH);
	}

	void orientizePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_ORIENTIZE);
	}

	void stickyPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((OctFrame) frame).callback(
			OctInterface.ACTION_STICKY);
	}

	void practicePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		practicePieces();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	void printPositions() {
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < sizeSize; position++)
			{
				int pos = startLoc[face][position].face;
				int rot = startLoc[face][position].rotation;

				if (pos < 100)
					System.out.print(" ");
				if (pos < 10 && pos >= 0)
					System.out.print(" ");
				System.out.print(pos + "," + rot);
				if (position % size == size - 1) {
					System.out.println();
				} else {
					System.out.print(" ");
				}
			}
			System.out.println();
		}
	}

	void printFacets() {
		for (int face = 0; face < MAX_SIDES; face++) {
			int square_1 = 0, sqrt = 1;

			for (int position = 0; position < sizeSize;
					position++) {
				System.out.print(
					facetLoc[face][position].face + " " +
					facetLoc[face][position].rotation +
					"  ");
				if (position == square_1) {
					System.out.println();
					++sqrt;
					square_1 = sqrt * sqrt - 1;
				}
			}
			System.out.println();
		}
		System.out.println();
	}

	void printFace() {
		int square_1 = 0, sqrt = 1;

		for (int position = 0; position < sizeSize;
				position++) {
			System.out.print(
				faceLoc[position].face + " " +
				faceLoc[position].rotation +
				"  ");
			if (position == square_1) {
				System.out.println();
				++sqrt;
				square_1 = sqrt * sqrt - 1;
			}
		}
		System.out.println();
	}


	void printRow(int len, int orientLine) {
		System.out.println("Row " + orientLine + ":");
		for (int i = 0; i < len; i++)
			System.out.print(
				rowLoc[orientLine][i].face + " " +
				rowLoc[orientLine][i].rotation +
				"  ");
		System.out.println();
	}


	void getStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeSize; position++) {
				startLoc[face][position].face =
					facetLoc[face][position].face;
				startLoc[face][position].rotation =
					facetLoc[face][position].rotation;
			}
	}

	void setStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeSize; position++) {
				facetLoc[face][position].face =
					startLoc[face][position].face;
				facetLoc[face][position].rotation =
					startLoc[face][position].rotation;
			}
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int group = (orient) ? 2 : 1;

			++line; //read startingPosition:
			for (int face = 0; face < MAX_FACES; face++) {
				int square_1 = 0, sqrt = 1, pos = 0;
				 ++line;
				while (token.length > line && token[line].equals(""))
					++line;
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int position = 0; position < sizeSize; position++) {
					int num = Integer.parseInt(st[group * pos + offset].trim());

					startLoc[face][position].face = num;
					if (orient) {
						num = Integer.parseInt(st[group * pos + offset + 1].trim());
						startLoc[face][position].rotation = num;
					}
					if (position == square_1) {
						if (position < sizeSize - 1) {
							line++;
							if (line >= token.length) {
								System.out.println("Short data in StartPosition");
								return -1;
							}
							lineData = token[line];
							st = lineData.split(space);
						}
						pos = 0;
						++sqrt;
						square_1 = sqrt * sqrt - 1;
					} else
						++pos;
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found in StartPosition");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				String[] st = lineData.substring(
					lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentStyle = Integer.parseInt(st[1].trim());
				currentControl = (Integer.parseInt(st[2].trim()) != 0);
				currentFace = Integer.parseInt(st[3].trim());
				currentPosition = Integer.parseInt(st[4].trim());
				{
					scanPaint = true;
					undo.setMove(currentDirection,
						currentStyle,
						(currentControl) ? 1 : 0,
						currentFace, currentPosition);
					redo.flushMoves();
					paint(g);
					if (!currentControl)
						((OctFrame) frame).callback(
							OctInterface.ACTION_MOVED);
				} /* else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
						return false;
				} */
			}
			if (l < moves) {
				showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Corrupt input in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int face = 0; face < MAX_FACES; face++) {
			int square_1 = 0, sqrt = 1;
			for (int position = 0; position < sizeSize; position++)
			{
				int temp = startLoc[face][position].face;
				if (temp < 100)
					sb.append(space);
				if (temp < 10 && temp >= 0)
					sb.append(space);
				sb.append(space).append(temp);
				if (orient) {
					temp = startLoc[face][position].rotation;
					if (temp < 10 && temp >= 0)
						sb.append(space);
					sb.append(space).append(temp);
				}
				if (position == square_1) {
					sb.append(newline);
					++sqrt;
					square_1 = sqrt * sqrt - 1;
				}
			}
			sb.append(newline);
		}
	}

	void printMoves(StringBuilder sb) {
		int direction, style, control;
		int face, position;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space).append("dir")
			.append(space).append("style")
			.append(space).append("control")
			.append(space).append("face")
			.append(space).append("position").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			style = undo.readMoveStyle();
			control = undo.readMoveControl();
			face = undo.readMoveFace();
			position = undo.readMovePosition();
			redo.setMove(direction, style, control, face,
				position);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			style = redo.readMoveStyle();
			control = redo.readMoveControl();
			face = redo.readMoveFace();
			position = redo.readMovePosition();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(space)
				.append(style).append(space)
				.append(control).append(space)
				.append(face).append(space)
				.append(position).append(newline);
			undo.setMove(direction, style, control, face,
				position);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (currentFace != -1) {
				drawTriangle(currentFace, currentPosition, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (currentFace != -1) {
				drawTriangle(currentFace, currentPosition, 0);
				selectPieces();
				currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			if (currentControl)
				moveControlCb(currentFace, currentDirection,
					currentStyle);
			else {
				movePieces(currentFace, toRTT(currentPosition),
					currentDirection, currentStyle);
				if (stackPaint && checkSolved()) {
					((OctFrame) frame).callback(
						OctInterface.ACTION_SOLVED);
				}
			}
			currentDirection = currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}

/**
 * The <code>OctLoc</code> class maintains the locations of
 * the pieces.
 */

class OctLoc {
	int	 face;
	int	 rotation;

	public OctLoc(int face, int rotation) {
		this.face = face;
		this.rotation = rotation;
	}
}

/**
 * The <code>OctRTT</code> class keeps track of row and diagonal
 * columns.
 */

class OctRTT {
	int	 row;
	int	 trbl;
	int	 tlbr;

	public OctRTT(int row, int trbl, int tlbr) {
		this.row = row;
		this.trbl = trbl;
		this.tlbr = tlbr;
	}
}

/**
 * The <code>OctNextRowP3</code> class maintains positional
 * information for rows for period 3 moves.
 */

class OctNextRowP3 {
	boolean     viewChanged;
	int	 face;
	int	 direction;

	public OctNextRowP3(boolean viewChanged, int face, int direction) {
		this.viewChanged = viewChanged;
		this.face = face;
		this.direction = direction;
	}
}
